# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests each of the functions in mongo_doc_manager
"""

import sys

sys.path[0:0] = [""]  # noqa

from mongo_connector.namespace_config import NamespaceConfig
from mongo_connector.command_helper import CommandHelper
from mongo_connector.doc_managers.mongo_doc_manager import DocManager
from mongo_connector.test_utils import TESTARGS, MockGridFSFile
from tests import unittest
from tests.test_mongo import MongoTestCase


class TestMongoDocManager(MongoTestCase):
    """Test class for MongoDocManager
    """

    id_field = "_id"

    @classmethod
    def setUpClass(cls):
        MongoTestCase.setUpClass()
        cls.namespaces_inc = ["test.test_include1", "test.test_include2"]

    def setUp(self):
        """Empty Mongo at the start of every test
        """

        self.choosy_docman = DocManager(
            self.standalone.uri,
            use_single_meta_collection=self.use_single_meta_collection,
        )

        self.mongo_conn.drop_database("__mongo_connector")
        self._remove()

        conn = self.standalone.client()
        for ns in self.namespaces_inc:
            db, coll = ns.split(".", 1)
            conn[db][coll].drop()

    def test_meta_collections(self):
        """Ensure that a DocManager returns the correct set of meta collection
        names.
        """
        meta_collection_names = set()
        if self.use_single_meta_collection:
            meta_collection_names.add(self.choosy_docman.meta_collection_name)
        # Before replication only the single meta_collection should be
        # returned.
        self.assertEqual(
            set(self.choosy_docman._meta_collections()), meta_collection_names
        )
        for namespace in ["test.test", "foo.bar", "test.bar"]:
            if not self.use_single_meta_collection:
                meta_collection_names.add(namespace)
            self.choosy_docman.upsert({"_id": "1"}, namespace, 1)
            self.assertEqual(
                set(self.choosy_docman._meta_collections()), meta_collection_names
            )

    def test_update(self):
        doc_id = "1"
        doc = {"_id": doc_id, "a": 1, "b": 2}
        self.mongo.insert_one(doc)
        # $set only
        update_spec = {"$set": {"a": 1, "b": 2}}
        doc = self.choosy_docman.update(doc_id, update_spec, *TESTARGS)
        self.assertEqual(doc, {"_id": doc_id, "a": 1, "b": 2})
        # $unset only
        update_spec = {"$unset": {"a": True}}
        doc = self.choosy_docman.update(doc_id, update_spec, *TESTARGS)
        self.assertEqual(doc, {"_id": doc_id, "b": 2})
        # mixed $set/$unset
        update_spec = {"$unset": {"b": True}, "$set": {"c": 3}}
        doc = self.choosy_docman.update(doc_id, update_spec, *TESTARGS)
        self.assertEqual(doc, {"_id": doc_id, "c": 3})

    def test_upsert(self):
        """Ensure we can properly insert into Mongo via DocManager.
        """

        docc = {"_id": "1", "name": "John"}
        self.choosy_docman.upsert(docc, *TESTARGS)
        res = list(self._search())
        self.assertEqual(len(res), 1)
        for doc in res:
            self.assertEqual(doc["_id"], "1")
            self.assertEqual(doc["name"], "John")

        docc = {"_id": "1", "name": "Paul"}
        self.choosy_docman.upsert(docc, *TESTARGS)
        res = list(self._search())
        self.assertEqual(len(res), 1)
        for doc in res:
            self.assertEqual(doc["_id"], "1")
            self.assertEqual(doc["name"], "Paul")

    def test_bulk_upsert(self):
        """Test the bulk_upsert method."""
        docs = ({"_id": i} for i in range(1000))
        self.choosy_docman.bulk_upsert(docs, *TESTARGS)
        res = list(self._search(sort=[("_id", 1)]))
        self.assertEqual(len(res), 1000)
        for i, r in enumerate(res):
            self.assertEqual(r["_id"], i)

        docs = ({"_id": i, "weight": 2 * i} for i in range(1000))
        self.choosy_docman.bulk_upsert(docs, *TESTARGS)

        res = list(self._search(sort=[("_id", 1)]))
        self.assertEqual(len(res), 1000)
        for i, r in enumerate(res):
            self.assertEqual(r["weight"], 2 * i)

    def test_remove(self):
        """Ensure we can properly delete from Mongo via DocManager.
        """

        docc = {"_id": "1", "name": "John"}
        self.choosy_docman.upsert(docc, *TESTARGS)
        self.assertEqual(len(list(self._search())), 1)
        self.choosy_docman.remove(docc["_id"], *TESTARGS)
        self.assertEqual(len(list(self._search())), 0)

    def test_insert_file(self):
        # Drop database, so that mongo_doc's client refreshes its index cache.
        self.choosy_docman.mongo.drop_database("test")
        test_data = " ".join(str(x) for x in range(100000)).encode("utf8")
        docc = {
            "_id": "test_id",
            "filename": "test_filename",
            "upload_date": 5,
            "md5": "test_md5",
        }
        self.choosy_docman.insert_file(MockGridFSFile(docc, test_data), *TESTARGS)
        res = self._search()
        for doc in res:
            self.assertEqual(doc[self.id_field], docc["_id"])
            self.assertEqual(doc["filename"], docc["filename"])
            self.assertEqual(doc["content"], test_data)

    def test_remove_file(self):
        # Drop database, so that mongo_doc's client refreshes its index cache.
        self.choosy_docman.mongo.drop_database("test")
        test_data = b"hello world"
        docc = {
            "_id": "test_id",
            "filename": "test_filename",
            "upload_date": 5,
            "md5": "test_md5",
        }

        self.choosy_docman.insert_file(MockGridFSFile(docc, test_data), *TESTARGS)
        res = list(self._search())
        self.assertEqual(len(res), 1)

        self.choosy_docman.remove(docc["_id"], *TESTARGS)
        res = list(self._search())
        self.assertEqual(len(res), 0)

    def test_search(self):
        """Query Mongo for docs in a timestamp range.

        We use API and DocManager's search(start_ts,end_ts), and then compare.
        """

        docc = {"_id": "1", "name": "John"}
        self.choosy_docman.upsert(docc, "test.test_include1", 5767301236327972865)
        docc2 = {"_id": "2", "name": "John Paul"}
        self.choosy_docman.upsert(docc2, "test.test_include1", 5767301236327972866)
        docc3 = {"_id": "3", "name": "Paul"}
        self.choosy_docman.upsert(docc3, "test.test_include1", 5767301236327972870)
        search = list(
            self.choosy_docman.search(5767301236327972865, 5767301236327972866)
        )
        self.assertEqual(len(search), 2)
        result_id = [result.get(self.id_field) for result in search]
        self.assertIn("1", result_id)
        self.assertIn("2", result_id)

    def test_search_namespaces(self):
        """Test search within timestamp range with a given namespace set
        """

        for ns in self.namespaces_inc:
            for i in range(100):
                self.choosy_docman.upsert({"_id": i}, ns, i)

        results = list(self.choosy_docman.search(0, 49))
        self.assertEqual(len(results), 100)
        for r in results:
            self.assertGreaterEqual(r[self.id_field], 0)

    def test_get_last_doc(self):
        """Insert documents, verify that get_last_doc() returns the one with
            the latest timestamp.
        """
        docc = {"_id": "4", "name": "Hare"}
        self.choosy_docman.upsert(docc, "test.test_include1", 3)
        docc = {"_id": "5", "name": "Tortoise"}
        self.choosy_docman.upsert(docc, "test.test_include1", 2)
        docc = {"_id": "6", "name": "Mr T."}
        self.choosy_docman.upsert(docc, "test.test_include1", 1)
        doc = self.choosy_docman.get_last_doc()
        self.assertEqual(doc[self.id_field], "4")
        docc = {"_id": "6", "name": "HareTwin"}
        self.choosy_docman.upsert(docc, "test.test_include1", 4)
        doc = self.choosy_docman.get_last_doc()
        self.assertEqual(doc[self.id_field], "6")

    def test_commands(self):
        # Also test with namespace mapping.
        # Note that mongo-connector does not currently support commands after
        # renaming a database.
        namespace_config = NamespaceConfig(
            namespace_set=["test.test", "test.test2", "test.drop"],
            namespace_options={
                "test.test": "test.othertest",
                "test.drop": "dropped.collection",
            },
        )
        self.choosy_docman.command_helper = CommandHelper(namespace_config)

        try:
            self.choosy_docman.handle_command({"create": "test"}, *TESTARGS)
            self.assertIn("othertest", self.mongo_conn["test"].collection_names())
            self.choosy_docman.handle_command(
                {"renameCollection": "test.test", "to": "test.test2"}, "admin.$cmd", 1
            )
            self.assertNotIn("othertest", self.mongo_conn["test"].collection_names())
            self.assertIn("test2", self.mongo_conn["test"].collection_names())

            self.choosy_docman.handle_command({"drop": "test2"}, "test.$cmd", 1)
            self.assertNotIn("test2", self.mongo_conn["test"].collection_names())

            # WiredTiger drops the database when the last collection is
            # dropped.
            if "test" not in self.mongo_conn.database_names():
                self.choosy_docman.handle_command({"create": "test"}, *TESTARGS)
            self.assertIn("test", self.mongo_conn.database_names())
            self.choosy_docman.handle_command({"dropDatabase": 1}, "test.$cmd", 1)
            self.assertNotIn("test", self.mongo_conn.database_names())

            # Briefly test mapped database name with dropDatabase command.
            self.mongo_conn.dropped.collection.insert_one({"a": 1})
            self.assertIn("dropped", self.mongo_conn.database_names())
            self.choosy_docman.handle_command({"dropDatabase": 1}, "test.$cmd", 1)
            self.assertNotIn("dropped", self.mongo_conn.database_names())
        finally:
            self.mongo_conn.drop_database("test")


class TestMongoDocManagerWithSingleMetaCollection(TestMongoDocManager):
    id_field = "doc_id"
    use_single_meta_collection = True


if __name__ == "__main__":
    unittest.main()



# Copyright 2016 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re
import sys

import importlib_resources

from tests import unittest
from mongo_connector.namespace_config import (
    NamespaceConfig,
    Namespace,
    match_replace_regex,
    namespace_to_regex,
    RegexSet,
    wildcards_overlap,
)
from mongo_connector import errors


class TestNamespaceConfig(unittest.TestCase):
    """Test the NamespaceConfig class"""

    def test_default(self):
        """Test that by default, all namespaces are kept without renaming"""
        namespace_config = NamespaceConfig()
        self.assertEqual(namespace_config.unmap_namespace("db1.col1"), "db1.col1")
        self.assertEqual(namespace_config.map_db("db1"), ["db1"])
        self.assertEqual(namespace_config.map_namespace("db1.col1"), "db1.col1")

    def test_config(self):
        """Test that the namespace option in the example config is valid."""
        package = "mongo_connector.service"
        stream = importlib_resources.open_text(package, "config.json")
        with stream:
            namespaces = json.load(stream)["__namespaces"]
        NamespaceConfig(namespace_options=namespaces)

    def test_include_plain(self):
        """Test including namespaces without wildcards"""
        namespace_config = NamespaceConfig(namespace_set=["db1.col1", "db1.col2"])
        self.assertEqual(namespace_config.unmap_namespace("db1.col1"), "db1.col1")
        self.assertEqual(namespace_config.unmap_namespace("db1.col2"), "db1.col2")
        self.assertIsNone(namespace_config.unmap_namespace("not.included"))
        self.assertEqual(namespace_config.map_db("db1"), ["db1"])
        self.assertEqual(namespace_config.map_db("not_included"), [])
        self.assertEqual(namespace_config.map_namespace("db1.col1"), "db1.col1")
        self.assertEqual(namespace_config.map_namespace("db1.col2"), "db1.col2")
        self.assertIsNone(namespace_config.map_namespace("db1.col4"))

    def test_include_wildcard(self):
        """Test including namespaces with wildcards"""
        equivalent_namespace_configs = (
            NamespaceConfig(namespace_set=["db1.*"]),
            NamespaceConfig(namespace_options={"db1.*": {}}),
            NamespaceConfig(namespace_options={"db1.*": True}),
            NamespaceConfig(namespace_options={"db1.*": {"rename": "db1.*"}}),
        )
        for namespace_config in equivalent_namespace_configs:
            self.assertEqual(namespace_config.unmap_namespace("db1.col1"), "db1.col1")
            self.assertEqual(namespace_config.unmap_namespace("db1.col2"), "db1.col2")
            self.assertEqual(
                namespace_config.lookup("db1.col1"),
                Namespace(dest_name="db1.col1", source_name="db1.col1"),
            )
            self.assertListEqual(namespace_config.map_db("db1"), ["db1"])
            self.assertEqual(namespace_config.map_namespace("db1.col1"), "db1.col1")
            self.assertIsNone(namespace_config.map_namespace("db2.col4"))

    def test_map_db_wildcard(self):
        """Test a crazy namespace renaming scheme with wildcards."""
        namespace_config = NamespaceConfig(
            namespace_options={
                "db.1_*": "db1.new_*",
                "db.2_*": "db2.new_*",
                "db.3": "new_db.3",
            }
        )
        self.assertEqual(
            set(namespace_config.map_db("db")), set(["db1", "db2", "new_db"])
        )

    def test_include_wildcard_periods(self):
        """Test the '.' in the namespace only matches '.'"""
        namespace_config = NamespaceConfig(namespace_set=["db.*"])
        self.assertIsNone(namespace_config.map_namespace("dbxcol"))
        self.assertEqual(namespace_config.map_namespace("db.col"), "db.col")

    def test_include_wildcard_multiple_periods(self):
        """Test matching a namespace with multiple '.' characters."""
        namespace_config = NamespaceConfig(namespace_set=["db.col.*"])
        self.assertIsNone(namespace_config.map_namespace("db.col"))
        self.assertEqual(namespace_config.map_namespace("db.col."), "db.col.")

    def test_include_wildcard_no_period_in_database(self):
        """Test that a database wildcard cannot match a period."""
        namespace_config = NamespaceConfig(namespace_set=["db*.col"])
        self.assertIsNone(namespace_config.map_namespace("db.bar.col"))
        self.assertEqual(namespace_config.map_namespace("dbfoo.col"), "dbfoo.col")

    def test_include_wildcard_metacharacters(self):
        """Test namespaces with metacharacters are matched."""
        namespace_config = NamespaceConfig(namespace_set=["db&_*.$_^_#_!_[_]_"])
        self.assertEqual(
            namespace_config.map_namespace("db&_foo.$_^_#_!_[_]_"),
            "db&_foo.$_^_#_!_[_]_",
        )
        self.assertIsNone(namespace_config.map_namespace("db&.foo"))

    def test_gridfs(self):
        """Test the gridfs property is set correctly."""
        equivalent_namespace_configs = (
            NamespaceConfig(gridfs_set=["db1.*"]),
            NamespaceConfig(namespace_options={"db1.*": {"gridfs": True}}),
        )
        for namespace_config in equivalent_namespace_configs:
            self.assertEqual(namespace_config.unmap_namespace("db1.col1"), "db1.col1")
            self.assertEqual(namespace_config.unmap_namespace("db1.col2"), "db1.col2")
            self.assertEqual(
                namespace_config.lookup("db1.col1"),
                Namespace(dest_name="db1.col1", source_name="db1.col1", gridfs=True),
            )
            self.assertListEqual(namespace_config.map_db("db1"), ["db1"])
            self.assertEqual(namespace_config.map_namespace("db1.col1"), "db1.col1")
            self.assertIsNone(namespace_config.map_namespace("db2.col4"))
            self.assertTrue(namespace_config.lookup("db1.col1").gridfs)
            self.assertEqual(namespace_config.gridfs_namespace("db1.col1"), "db1.col1")
            self.assertIsNone(namespace_config.gridfs_namespace("not.gridfs"))

    def test_exclude_plain(self):
        """Test excluding namespaces without wildcards"""
        namespace_config = NamespaceConfig(ex_namespace_set=["ex.clude"])
        self.assertEqual(namespace_config.unmap_namespace("db.col"), "db.col")
        self.assertEqual(namespace_config.unmap_namespace("ex.clude"), "ex.clude")
        self.assertEqual(namespace_config.map_namespace("db.col"), "db.col")
        self.assertIsNone(namespace_config.map_namespace("ex.clude"))

    def test_exclude_wildcard(self):
        """Test excluding namespaces with wildcards"""
        equivalent_namespace_configs_for_tests = (
            NamespaceConfig(ex_namespace_set=["ex.*", "ex2.*"]),
            NamespaceConfig(namespace_options={"ex.*": False, "ex2.*": False}),
            # Multiple wildcards in exclude namespace
            NamespaceConfig(ex_namespace_set=["e*.*"]),
        )
        for namespace_config in equivalent_namespace_configs_for_tests:
            self.assertEqual(namespace_config.unmap_namespace("db.col"), "db.col")
            self.assertEqual(namespace_config.unmap_namespace("ex.clude"), "ex.clude")
            self.assertEqual(namespace_config.map_namespace("db.col"), "db.col")
            self.assertIsNone(namespace_config.map_namespace("ex.clude"))
            self.assertIsNone(namespace_config.map_namespace("ex2.clude"))

    def test_include_and_exclude(self):
        """Test including and excluding namespaces at the same time."""
        equivalent_namespace_configs_for_tests = (
            NamespaceConfig(
                ex_namespace_set=["ex.*"],
                namespace_set=["ex.cluded_still", "in.cluded"],
            ),
            NamespaceConfig(
                namespace_options={
                    "ex.*": False,
                    "ex.cluded_still": True,
                    "in.cluded": True,
                }
            ),
            NamespaceConfig(
                ex_namespace_set=["ex.cluded", "ex.cluded_still"],
                namespace_set=["ex.*", "in.cluded"],
            ),
        )
        for namespace_config in equivalent_namespace_configs_for_tests:
            self.assertIsNone(namespace_config.map_namespace("ex.cluded"))
            # Excluded namespaces take precedence over included ones.
            self.assertIsNone(namespace_config.map_namespace("ex.cluded_still"))
            # Namespaces that are not explicitly included are ignored.
            self.assertIsNone(namespace_config.map_namespace("also.not.included"))
            self.assertEqual(namespace_config.map_namespace("in.cluded"), "in.cluded")

    def test_include_and_exclude_validation(self):
        """Test including and excluding the same namespaces is an error."""
        equivalent_namespace_config_kwargs = (
            dict(
                ex_namespace_set=["ex.cluded"], namespace_set=["in.cluded", "ex.cluded"]
            ),
            dict(
                namespace_set=["ex.cluded"],
                namespace_options={"ex.cluded": False, "in.cluded": True},
            ),
            dict(
                ex_namespace_set=["ex.cluded", "in.cluded"],
                namespace_options={"in.cluded": True},
            ),
        )
        for kwargs in equivalent_namespace_config_kwargs:
            with self.assertRaises(errors.InvalidConfiguration):
                NamespaceConfig(**kwargs)

    def test_unmap_namespace_wildcard(self):
        """Test un-mapping a namespace that was never explicitly mapped."""
        namespace_config = NamespaceConfig(
            namespace_options={"db2.*": "db2.f*", "db_*.foo": "db_new_*.foo"}
        )
        self.assertEqual(namespace_config.unmap_namespace("db2.foo"), "db2.oo")
        self.assertEqual(
            namespace_config.unmap_namespace("db_new_123.foo"), "db_123.foo"
        )

    def test_override_namespace_options(self):
        """Test gridfs_set and dest_mapping arguments override
        namespace_options.
        """
        namespace_config = NamespaceConfig(
            namespace_set=["override.me", "override.me2"],
            gridfs_set=["override.me3"],
            dest_mapping={
                "override.me": "overridden.1",
                "override.me2": "overridden.2",
            },
            namespace_options={
                "override.me": {
                    "rename": "override.me",
                    "includeFields": ["_id", "dont_remove"],
                },
                "override.me2": "override.me2",
                "override.me3": {"gridfs": False},
            },
        )
        overridden = namespace_config.lookup("override.me")
        self.assertEqual(overridden.dest_name, "overridden.1")
        self.assertEqual(overridden.include_fields, set(["_id", "dont_remove"]))
        overridden = namespace_config.lookup("override.me2")
        self.assertEqual(overridden.dest_name, "overridden.2")
        self.assertFalse(overridden.include_fields)
        self.assertTrue(namespace_config.gridfs_namespace("override.me3"))

    def test_invalid_collection_name_validation(self):
        """Test that invalid collection names raise InvalidConfiguration."""
        equivalent_namespace_config_kwargs = (
            dict(namespace_options={"invalid_db": "newinvalid_db"}),
            dict(namespace_set=["invalid_db."]),
            dict(ex_namespace_set=[".invalid_db"]),
            dict(gridfs_set=[".invalid_db"]),
        )
        for kwargs in equivalent_namespace_config_kwargs:
            with self.assertRaises(errors.InvalidConfiguration):
                NamespaceConfig(**kwargs)

    def test_gridfs_rename_invalid(self):
        """Test that renaming a GridFS collection is invalid."""
        with self.assertRaises(errors.InvalidConfiguration):
            NamespaceConfig(
                namespace_options={
                    "gridfs.*": {"rename": "new_gridfs.*", "gridfs": True}
                }
            )

    def test_rename_validation(self):
        """Test namespace renaming validation."""
        # Multiple collections cannot be merged into the same target namespace
        with self.assertRaises(errors.InvalidConfiguration):
            NamespaceConfig(
                namespace_options={
                    "db1.col1": "newdb.newcol",
                    "db2.col1": "newdb.newcol",
                }
            )

        # Multiple collections cannot be merged into the same target namespace
        with self.assertRaises(errors.InvalidConfiguration):
            NamespaceConfig(
                namespace_options={
                    "db*.col1": "newdb.newcol*",
                    "db*.col2": "newdb.newcol*",
                }
            )

        # Multiple collections cannot be merged into the same target namespace
        namespace_config = NamespaceConfig(
            namespace_options={"*.coll": "*.new_coll", "db.*": "new_db.*"}
        )
        namespace_config.map_namespace("new_db.coll")
        with self.assertRaises(errors.InvalidConfiguration):
            # "db.new_coll" should map to "new_db.new_coll" but there is
            # already a mapping from "new_db.coll" to "new_db.new_coll".
            namespace_config.map_namespace("db.new_coll")

        # For the sake of map_db, wildcards cannot be moved from database name
        # to collection name.
        with self.assertRaises(errors.InvalidConfiguration):
            NamespaceConfig(namespace_options={"db*.col": "new_db.col_*"})

        # For the sake of map_db, wildcards cannot be moved from collection
        # name to database name.
        with self.assertRaises(errors.InvalidConfiguration):
            NamespaceConfig(namespace_options={"db.*": "new_db_*.col"})

    def test_fields_validation(self):
        """Test including/excluding fields per namespace."""
        # Cannot include and exclude fields in the same namespace
        with self.assertRaises(errors.InvalidConfiguration):
            NamespaceConfig(
                namespace_options={
                    "db.col": {"includeFields": ["a"], "excludeFields": ["b"]}
                }
            )

        # Cannot include fields globally and then exclude fields
        with self.assertRaises(errors.InvalidConfiguration):
            NamespaceConfig(
                include_fields=["a"],
                namespace_options={"db.col": {"excludeFields": ["b"]}},
            )

        # Cannot exclude fields globally and then include fields
        with self.assertRaises(errors.InvalidConfiguration):
            NamespaceConfig(
                exclude_fields=["b"],
                namespace_options={"db.col": {"includeFields": ["a"]}},
            )

    def test_projection_include_wildcard(self):
        """Test include_fields on a wildcard namespace."""
        equivalent_namespace_configs = (
            NamespaceConfig(
                include_fields=["foo", "nested.field"],
                ex_namespace_set=["ignored.name"],
            ),
            NamespaceConfig(
                include_fields=["foo", "nested.field"], namespace_set=["db.foo"]
            ),
            NamespaceConfig(
                include_fields=["foo", "nested.field"], namespace_set=["db.*"]
            ),
            NamespaceConfig(
                namespace_options={"db.*": {"includeFields": ["foo", "nested.field"]}}
            ),
            NamespaceConfig(
                namespace_options={"db.foo": {"includeFields": ["foo", "nested.field"]}}
            ),
            NamespaceConfig(
                include_fields=["foo", "nested.field"],
                namespace_options={"db.*": {"includeFields": ["foo", "nested.field"]}},
            ),
        )
        for namespace_config in equivalent_namespace_configs:
            self.assertEqual(
                namespace_config.projection("db.foo"),
                {"_id": 1, "foo": 1, "nested.field": 1},
            )
            self.assertIsNone(namespace_config.projection("ignored.name"))

    def test_projection_exclude_wildcard(self):
        """Test exclude_fields on a wildcard namespace."""
        equivalent_namespace_configs = (
            NamespaceConfig(
                exclude_fields=["_id", "foo", "nested.field"], namespace_set=["db.*"]
            ),
            NamespaceConfig(
                namespace_options={
                    "db.*": {"excludeFields": ["_id", "foo", "nested.field"]}
                }
            ),
            NamespaceConfig(
                exclude_fields=["foo", "nested.field"],
                namespace_options={
                    "db.*": {"excludeFields": ["_id", "foo", "nested.field"]}
                },
            ),
        )
        for namespace_config in equivalent_namespace_configs:
            self.assertEqual(
                namespace_config.projection("db.foo"), {"foo": 0, "nested.field": 0}
            )
            self.assertIsNone(namespace_config.projection("ignored.name"))

    def test_get_included_databases(self):
        """Test get_included_databases."""
        nc_requires_list_databases = (
            NamespaceConfig(),
            NamespaceConfig(namespace_options={"db.c": False}),
            NamespaceConfig(namespace_options={"db*.c": True}),
        )
        for namespace_config in nc_requires_list_databases:
            self.assertEqual(namespace_config.get_included_databases(), [])

        namespace_config = NamespaceConfig(namespace_options={"db.c": True})
        self.assertEqual(namespace_config.get_included_databases(), ["db"])
        namespace_config = NamespaceConfig(namespace_options={"db.*": True})
        self.assertEqual(namespace_config.get_included_databases(), ["db"])

    def test_match_replace_regex(self):
        """Test regex matching and replacing."""
        regex = re.compile(r"\Adb_([^.]*).foo\Z")
        self.assertIsNone(match_replace_regex(regex, "db.foo", "*.foo"))
        self.assertIsNone(match_replace_regex(regex, "db.foo.foo", "*.foo"))
        self.assertEqual(match_replace_regex(regex, "db_bar.foo", "*.foo"), "bar.foo")

    def test_namespace_to_regex(self):
        """Test regex creation."""
        self.assertEqual(
            namespace_to_regex("db*.foo"), re.compile(r"\Adb([^.]*)\.foo\Z")
        )
        self.assertEqual(namespace_to_regex("db.foo*"), re.compile(r"\Adb\.foo(.*)\Z"))
        self.assertEqual(namespace_to_regex("db.foo"), re.compile(r"\Adb\.foo\Z"))

    @unittest.skipIf(sys.version_info >= (3, 7), "Python 3.6 and earlier")
    def test_namespace_to_regex_escapes_metacharacters_legacy(self):
        """Test regex creation escapes metacharacters."""
        self.assertEqual(
            namespace_to_regex("db&*.$a^a#a!a[a]a"),
            re.compile(r"\Adb\&([^.]*)\.\$a\^a\#a\!a\[a\]a\Z"),
        )
        self.assertEqual(
            namespace_to_regex("db.$a^a#a!a[a]a*"),
            re.compile(r"\Adb\.\$a\^a\#a\!a\[a\]a(.*)\Z"),
        )

    @unittest.skipIf(sys.version_info < (3, 7), "Python 3.7 and later")
    def test_namespace_to_regex_escapes_metacharacters(self):
        """Test regex creation escapes metacharacters."""
        self.assertEqual(
            namespace_to_regex("db&*.$a^a#a!a[a]a"),
            re.compile(r"\Adb\&([^.]*)\.\$a\^a\#a!a\[a\]a\Z"),
        )
        self.assertEqual(
            namespace_to_regex("db.$a^a#a!a[a]a*"),
            re.compile(r"\Adb\.\$a\^a\#a!a\[a\]a(.*)\Z"),
        )

    def test_wildcards_overlap(self):
        """Test wildcard strings """
        self.assertTrue(wildcards_overlap("foo.bar", "foo.bar"))
        self.assertTrue(wildcards_overlap("f*.bar", "foo.bar"))
        self.assertTrue(wildcards_overlap("*.bar", "foo.*"))
        self.assertTrue(wildcards_overlap("f*.bar", "foo.b*"))
        self.assertTrue(wildcards_overlap("a*b*c", "*c*"))
        self.assertFalse(wildcards_overlap("foo.bar", "foo.bar2"))
        self.assertFalse(wildcards_overlap("foo.*2", "foo.*1"))
        self.assertFalse(wildcards_overlap("foo.*", "food.*"))
        self.assertFalse(wildcards_overlap("a*b*c", "*c*d"))


class TestRegexSet(unittest.TestCase):
    """Test the RegexSet class."""

    def test_from_namespaces(self):
        """Test construction from list of namespaces."""
        self.assertEqual(RegexSet.from_namespaces([]), RegexSet([], []))
        self.assertEqual(
            RegexSet.from_namespaces(["db.bar", "db_*.foo", "db2.*"]),
            RegexSet(
                [namespace_to_regex("db_*.foo"), namespace_to_regex("db2.*")],
                ["db.bar"],
            ),
        )

    def test_contains(self):
        """Test membership query."""
        regex_set = RegexSet.from_namespaces(["db.bar", "db_*.foo", "db2.*"])
        self.assertTrue("db.bar" in regex_set)
        self.assertTrue("db_1.foo" in regex_set)
        self.assertTrue("db2.bar" in regex_set)
        self.assertTrue("db2.bar2" in regex_set)
        self.assertFalse("not.found" in regex_set)
        self.assertFalse("db_not.found" in regex_set)

    def test_add(self):
        """Test adding a new string."""
        regex_set = RegexSet.from_namespaces([])
        self.assertFalse("string.found" in regex_set)
        regex_set.add("string.found")
        self.assertTrue("string.found" in regex_set)

    def test_discard(self):
        """Test discarding a string."""
        regex_set = RegexSet.from_namespaces(["db.bar"])
        self.assertTrue("db.bar" in regex_set)
        regex_set.discard("db.bar")
        self.assertFalse("db.bar" in regex_set)


if __name__ == "__main__":
    unittest.main()



# Copyright 2016 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re

from collections import namedtuple, MutableSet
from itertools import combinations

from mongo_connector import errors


LOG = logging.getLogger(__name__)


_Namespace = namedtuple(
    "Namespace",
    ["dest_name", "source_name", "gridfs", "include_fields", "exclude_fields"],
)


class Namespace(_Namespace):
    def __new__(
        cls,
        dest_name=None,
        source_name=None,
        gridfs=False,
        include_fields=None,
        exclude_fields=None,
    ):
        include_fields = set(include_fields or [])
        exclude_fields = set(exclude_fields or [])
        return super(Namespace, cls).__new__(
            cls, dest_name, source_name, gridfs, include_fields, exclude_fields
        )

    def with_options(self, **kwargs):
        new_options = dict(
            dest_name=self.dest_name,
            source_name=self.source_name,
            gridfs=self.gridfs,
            include_fields=self.include_fields,
            exclude_fields=self.exclude_fields,
        )
        new_options.update(kwargs)
        return Namespace(**new_options)


class RegexSet(MutableSet):
    """Set that stores both plain strings and RegexObjects.

    Membership query results are cached so that repeated lookups of the same
    string are fast.
    """

    def __init__(self, regexes, strings):
        self._regexes = set(regexes)
        self._plain = set(strings)
        self._not_found_cache = set()

    def __contains__(self, item):
        if item in self._not_found_cache:
            return False
        if item in self._plain:
            return True
        if item in self._regexes:
            return True
        for regex in self._regexes:
            if regex.match(item):
                self._plain.add(item)
                return True
        self._not_found_cache.add(item)
        return False

    def __iter__(self):
        for regex in self._regexes:
            yield regex
        for string in self._plain:
            yield string

    def __len__(self):
        return len(self._regexes) + len(self._plain)

    def add(self, string):
        self._plain.add(string)
        self._not_found_cache.discard(string)

    def discard(self, string):
        self._plain.discard(string)

    @staticmethod
    def from_namespaces(namespaces):
        regexes = set()
        strings = set()
        for ns in namespaces:
            if "*" in ns:
                regexes.add(namespace_to_regex(ns))
            else:
                strings.add(ns)
        return RegexSet(regexes, strings)


class NamespaceConfig(object):
    """Manages included and excluded namespaces.
    """

    def __init__(
        self,
        namespace_set=None,
        ex_namespace_set=None,
        gridfs_set=None,
        dest_mapping=None,
        namespace_options=None,
        include_fields=None,
        exclude_fields=None,
    ):
        # A mapping from non-wildcard source namespaces to a MappedNamespace
        # containing the non-wildcard target name.
        self._plain = {}
        # A mapping from non-wildcard target namespaces to their
        # corresponding non-wildcard source namespace. Namespaces have a
        # one-to-one relationship with the target system, meaning multiple
        # source namespaces cannot be merged into a single namespace in the
        # target.
        self._reverse_plain = {}
        # A mapping from non-wildcard source database names to the set of
        # non-wildcard target database names.
        self._plain_db = {}
        # A list of (re.RegexObject, MappedNamespace) tuples. regex_map maps
        # wildcard source namespaces to a MappedNamespace containing the
        # wildcard target name. When a namespace is matched, an entry is
        # created in `self.plain` for faster subsequent lookups.
        self._regex_map = []

        # Fields to include or exclude from all namespaces
        self._include_fields = validate_include_fields(include_fields)
        self._exclude_fields = validate_exclude_fields(exclude_fields)

        # Add each included namespace. Namespaces have a one-to-one
        # relationship to the target system, meaning multiple source
        # namespaces cannot be merged into a single namespace in the target.
        ex_namespace_set, namespaces = validate_namespace_options(
            namespace_set=namespace_set,
            ex_namespace_set=ex_namespace_set,
            gridfs_set=gridfs_set,
            dest_mapping=dest_mapping,
            namespace_options=namespace_options,
            include_fields=include_fields,
            exclude_fields=exclude_fields,
        )

        # The set of, possibly wildcard, namespaces to exclude.
        self._ex_namespace_set = RegexSet.from_namespaces(ex_namespace_set)

        for namespace in namespaces:
            self._register_namespace_and_command(namespace)

    def _register_namespace_and_command(self, namespace):
        """Add a Namespace and the corresponding command namespace."""
        self._add_namespace(namespace)
        # Add the namespace for commands on this database
        cmd_name = namespace.source_name.split(".", 1)[0] + ".$cmd"
        dest_cmd_name = namespace.dest_name.split(".", 1)[0] + ".$cmd"
        self._add_namespace(Namespace(dest_name=dest_cmd_name, source_name=cmd_name))

    def _add_namespace(self, namespace):
        """Add an included and possibly renamed Namespace."""
        src_name = namespace.source_name
        if "*" in src_name:
            self._regex_map.append((namespace_to_regex(src_name), namespace))
        else:
            self._add_plain_namespace(namespace)

    def _add_plain_namespace(self, namespace):
        """Add an included and possibly renamed non-wildcard Namespace."""
        src_name = namespace.source_name
        target_name = namespace.dest_name
        src_names = self._reverse_plain.setdefault(target_name, set())
        src_names.add(src_name)
        if len(src_names) > 1:
            # Another source namespace is already mapped to this target
            existing_src = (src_names - set([src_name])).pop()
            raise errors.InvalidConfiguration(
                "Multiple namespaces cannot be combined into one target "
                "namespace. Trying to map '%s' to '%s' but there already "
                "exists a mapping from '%s' to '%s'"
                % (src_name, target_name, existing_src, target_name)
            )

        self._plain[src_name] = namespace
        src_db, _ = src_name.split(".", 1)
        target_db, _ = target_name.split(".", 1)
        self._plain_db.setdefault(src_db, set()).add(target_db)

    def lookup(self, plain_src_ns):
        """Given a plain source namespace, return the corresponding Namespace
        object, or None if it is not included.
        """
        # Ignore the namespace if it is excluded.
        if plain_src_ns in self._ex_namespace_set:
            return None
        # Include all namespaces if there are no included namespaces.
        if not self._regex_map and not self._plain:
            return Namespace(
                dest_name=plain_src_ns,
                source_name=plain_src_ns,
                include_fields=self._include_fields,
                exclude_fields=self._exclude_fields,
            )
        # First, search for the namespace in the plain namespaces.
        try:
            return self._plain[plain_src_ns]
        except KeyError:
            # Search for the namespace in the wildcard namespaces.
            for regex, namespace in self._regex_map:
                new_name = match_replace_regex(regex, plain_src_ns, namespace.dest_name)
                if not new_name:
                    continue
                # Save the new target Namespace in the plain namespaces so
                # future lookups are fast.
                new_namespace = namespace.with_options(
                    dest_name=new_name, source_name=plain_src_ns
                )
                self._add_plain_namespace(new_namespace)
                return new_namespace

        # Save the not included namespace to the excluded namespaces so
        # that future lookups of the same namespace are fast.
        self._ex_namespace_set.add(plain_src_ns)
        return None

    def map_namespace(self, plain_src_ns):
        """Given a plain source namespace, return the corresponding plain
        target namespace, or None if it is not included.
        """
        namespace = self.lookup(plain_src_ns)
        if namespace:
            return namespace.dest_name
        return None

    def gridfs_namespace(self, plain_src_ns):
        """Given a plain source namespace, return the corresponding plain
        target namespace if this namespace is a gridfs collection.
        """
        namespace = self.lookup(plain_src_ns)
        if namespace and namespace.gridfs:
            return namespace.dest_name
        return None

    def unmap_namespace(self, plain_target_ns):
        """Given a plain target namespace, return the corresponding source
        namespace.
        """
        # Return the same namespace if there are no included namespaces.
        if not self._regex_map and not self._plain:
            return plain_target_ns

        src_name_set = self._reverse_plain.get(plain_target_ns)
        if src_name_set:
            # Return the first (and only) item in the set
            for src_name in src_name_set:
                return src_name
        # The target namespace could also exist in the wildcard namespaces
        for _, namespace in self._regex_map:
            original_name = match_replace_regex(
                namespace_to_regex(namespace.dest_name),
                plain_target_ns,
                namespace.source_name,
            )
            if original_name:
                return original_name
        return None

    def map_db(self, plain_src_db):
        """Given a plain source database, return the list of target databases.

        Individual collections in a database can be mapped to different
        target databases, so map_db can return multiple databases. This
        function must return all target database names so we make the
        following restrictions on wildcards:
        1) A wildcard appearing in the source database name must also appear
        in the target database name, eg "db*.col" => "new_db_*.new_col".
        2) A wildcard appearing in the source collection name must also appear
        in the target collection name, eg "db.col*" => "new_db.new_col*".

        This is used by the CommandHelper for the dropDatabase command.
        """
        if not self._regex_map and not self._plain:
            return [plain_src_db]
        # Lookup this namespace to seed the plain_db dictionary
        self.lookup(plain_src_db + ".$cmd")
        return list(self._plain_db.get(plain_src_db, set()))

    def projection(self, plain_src_name):
        """Return the projection for the given source namespace."""
        mapped = self.lookup(plain_src_name)
        if not mapped:
            return None
        fields = mapped.include_fields or mapped.exclude_fields
        if fields:
            include = 1 if mapped.include_fields else 0
            return dict((field, include) for field in fields)
        return None

    def get_included_databases(self):
        """Return the databases we want to include, or empty list for all.
        """
        databases = set()
        databases.update(self._plain_db.keys())

        for _, namespace in self._regex_map:
            database_name, _ = namespace.source_name.split(".", 1)
            if "*" in database_name:
                return []
            databases.add(database_name)

        return list(databases)


def _character_matches(name1, name2):
    """Yield the number of characters that match the beginning of each string.
    """
    if name1[0] == "*":
        for i in range(len(name2) + 1):
            yield 1, i
    if name2[0] == "*":
        for i in range(len(name1) + 1):
            yield i, 1
    if name1[0] == name2[0]:
        yield 1, 1


def wildcards_overlap(name1, name2):
    """Return true if two wildcard patterns can match the same string."""
    if not name1 and not name2:
        return True
    if not name1 or not name2:
        return False
    for matched1, matched2 in _character_matches(name1, name2):
        if wildcards_overlap(name1[matched1:], name2[matched2:]):
            return True
    return False


def _validate_namespace(name):
    """Validate a MongoDB namespace."""
    if name.find(".", 1, len(name) - 1) < 0:
        raise errors.InvalidConfiguration("Invalid MongoDB namespace '%s'!" % (name,))


def _validate_namespaces(namespaces):
    """Validate wildcards and renaming in namespaces.

    Target namespaces should have the same number of wildcards as the source.
    No target namespaces overlap exactly with each other. Logs a warning
    when wildcard namespaces have a chance of overlapping.
    """
    for source, namespace in namespaces.items():
        target = namespace.dest_name
        _validate_namespace(source)
        _validate_namespace(target)
        if source.count("*") > 1 or target.count("*") > 1:
            raise errors.InvalidConfiguration(
                "The namespace mapping from '%s' to '%s' cannot contain more "
                "than one '*' character." % (source, target)
            )
        if source.count("*") != target.count("*"):
            raise errors.InvalidConfiguration(
                "The namespace mapping from '%s' to '%s' must contain the "
                "same number of '*' characters." % (source, target)
            )
        if "*" not in source:
            continue
        # Make sure that wildcards are not moved from database name to
        # collection name or vice versa, eg "db*.foo" => "db.foo_*"
        if (
            wildcard_in_db(source)
            and not wildcard_in_db(target)
            or (not wildcard_in_db(source) and wildcard_in_db(target))
        ):
            raise errors.InvalidConfiguration(
                "The namespace mapping from '%s' to '%s' is invalid. A '*' "
                "that appears in the source database name must also appear"
                "in the target database name. A '*' that appears in the "
                "source collection name must also appear in the target "
                "collection name" % (source, target)
            )

    for source1, source2 in combinations(namespaces.keys(), 2):
        if wildcards_overlap(source1, source2):
            LOG.warning(
                'Namespaces "%s" and "%s" may match the ' "same source namespace.",
                source1,
                source2,
            )
        target1 = namespaces[source1].dest_name
        target2 = namespaces[source2].dest_name
        if target1 == target2:
            raise errors.InvalidConfiguration(
                "Multiple namespaces cannot be combined into one target "
                "namespace. Trying to map '%s' to '%s' but '%s' already "
                "corresponds to '%s' in the target system."
                % (source2, target2, source1, target1)
            )
        if wildcards_overlap(target1, target2):
            LOG.warning(
                "Multiple namespaces cannot be combined into one target "
                "namespace. Mapping from '%s' to '%s' might overlap "
                "with mapping from '%s' to '%s'." % (source2, target2, source1, target1)
            )


def _merge_namespace_options(
    namespace_set=None,
    ex_namespace_set=None,
    gridfs_set=None,
    dest_mapping=None,
    namespace_options=None,
    include_fields=None,
    exclude_fields=None,
):
    """Merges namespaces options together.

    The first is the set of excluded namespaces and the second is a mapping
    from source namespace to Namespace instances.
    """
    namespace_set = set(namespace_set or [])
    ex_namespace_set = set(ex_namespace_set or [])
    gridfs_set = set(gridfs_set or [])
    dest_mapping = dest_mapping or {}
    namespace_options = namespace_options or {}
    include_fields = set(include_fields or [])
    exclude_fields = set(exclude_fields or [])
    namespaces = {}

    for source_name, options_or_str in namespace_options.items():
        if isinstance(options_or_str, dict):
            namespace_set.add(source_name)
            if options_or_str.get("gridfs"):
                gridfs_set.add(source_name)
            namespaces[source_name] = Namespace(
                dest_name=options_or_str.get("rename"),
                include_fields=options_or_str.get("includeFields"),
                exclude_fields=options_or_str.get("excludeFields"),
                gridfs=options_or_str.get("gridfs", False),
            )
        elif isinstance(options_or_str, str):
            namespace_set.add(source_name)
            namespaces[source_name] = Namespace(dest_name=options_or_str)
        elif options_or_str:
            namespace_set.add(source_name)
        else:
            ex_namespace_set.add(source_name)

    # Add namespaces that are renamed but not in namespace_options
    for source_name, target_name in dest_mapping.items():
        namespaces[source_name] = namespaces.get(source_name, Namespace()).with_options(
            dest_name=target_name
        )

    # Add namespaces that are included but not in namespace_options
    for included_name in namespace_set:
        if included_name not in namespaces:
            namespaces[included_name] = Namespace()

    # Add namespaces that are excluded but not in namespace_options
    for gridfs_name in gridfs_set:
        namespaces[gridfs_name] = namespaces.get(gridfs_name, Namespace()).with_options(
            gridfs=True
        )

    # Add source, destination name, and globally included and excluded fields
    for included_name in namespaces:
        namespace = namespaces[included_name]
        namespace = namespace.with_options(
            source_name=included_name,
            include_fields=validate_include_fields(
                include_fields, namespace.include_fields
            ),
            exclude_fields=validate_exclude_fields(
                exclude_fields, namespace.exclude_fields
            ),
        )
        # The default destination name is the same as the source.
        if not namespace.dest_name:
            namespace = namespace.with_options(dest_name=included_name)
        namespaces[included_name] = namespace

    return ex_namespace_set, namespaces


def validate_namespace_options(
    namespace_set=None,
    ex_namespace_set=None,
    gridfs_set=None,
    dest_mapping=None,
    namespace_options=None,
    include_fields=None,
    exclude_fields=None,
):
    ex_namespace_set, namespaces = _merge_namespace_options(
        namespace_set=namespace_set,
        ex_namespace_set=ex_namespace_set,
        gridfs_set=gridfs_set,
        dest_mapping=dest_mapping,
        namespace_options=namespace_options,
        include_fields=include_fields,
        exclude_fields=exclude_fields,
    )

    for excluded_name in ex_namespace_set:
        _validate_namespace(excluded_name)
        if excluded_name in namespaces:
            raise errors.InvalidConfiguration(
                "Cannot include namespace '%s', it is already excluded."
                % (excluded_name,)
            )

    for namespace in namespaces.values():
        if namespace.include_fields and namespace.exclude_fields:
            raise errors.InvalidConfiguration(
                "Cannot mix include fields and exclude fields in "
                "namespace mapping for: '%s'" % (namespace.source_name,)
            )

        if namespace.gridfs and namespace.dest_name != namespace.source_name:
            raise errors.InvalidConfiguration(
                "GridFS namespaces cannot be renamed: '%s'" % (namespace.source_name,)
            )

    _validate_namespaces(namespaces)
    return ex_namespace_set, namespaces.values()


def match_replace_regex(regex, src_namespace, dest_namespace):
    """Return the new mapped namespace if the src_namespace matches the
    regex."""
    match = regex.match(src_namespace)
    if match:
        return dest_namespace.replace("*", match.group(1))
    return None


def wildcard_in_db(namespace):
    """Return True if a wildcard character appears in the database name."""
    return namespace.find("*") < namespace.find(".")


def namespace_to_regex(namespace):
    """Create a RegexObject from a wildcard namespace."""
    db_name, coll_name = namespace.split(".", 1)
    # A database name cannot contain a '.' character
    db_regex = re.escape(db_name).replace(r"\*", "([^.]*)")
    # But a collection name can.
    coll_regex = re.escape(coll_name).replace(r"\*", "(.*)")
    return re.compile(r"\A" + db_regex + r"\." + coll_regex + r"\Z")


def validate_include_fields(include_fields, namespace_fields=None):
    include_fields = set(include_fields or [])
    namespace_fields = set(namespace_fields or [])
    merged = include_fields | namespace_fields
    if merged:
        merged.add("_id")
    return merged


def validate_exclude_fields(exclude_fields, namespace_fields=None):
    exclude_fields = set(exclude_fields or [])
    namespace_fields = set(namespace_fields or [])
    merged = exclude_fields | namespace_fields
    if "_id" in merged:
        LOG.warning("Cannot exclude '_id' field, ignoring")
        merged.discard("_id")
    return merged



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from setuptools import setup

classifiers = """\
Development Status :: 4 - Beta
Intended Audience :: Developers
License :: OSI Approved :: Apache Software License
Programming Language :: Python :: 3
Topic :: Database
Topic :: Software Development :: Libraries :: Python Modules
Operating System :: Unix
Operating System :: MacOS :: MacOS X
Operating System :: Microsoft :: Windows
Operating System :: POSIX
"""

setup(
    name="mongo-connector",
    use_scm_version=True,
    author="MongoDB, Inc.",
    author_email="mongodb-user@googlegroups.com",
    description="Mongo Connector",
    keywords=["mongo-connector", "mongo", "mongodb", "solr", "elasticsearch"],
    url="https://github.com/yougov/mongo-connector",
    platforms=["any"],
    classifiers=filter(None, classifiers.split("\n")),
    install_requires=[
        "pymongo >= 2.9",
        "importlib_metadata>=0.6",
        "autocommand",
        "importlib_resources",
    ],
    packages=["mongo_connector", "mongo_connector.doc_managers"],
    package_data={"mongo_connector.doc_managers": ["schema.xml"]},
    entry_points={
        "console_scripts": ["mongo-connector = mongo_connector.connector:main"]
    },
    extras_require={
        "solr": ["solr-doc-manager"],
        "elastic": ["elastic-doc-manager"],
        "elastic-aws": ["elastic-doc-manager[aws]"],
        "elastic2": ["elastic2-doc-manager[elastic2]"],
        "elastic5": ["elastic2-doc-manager[elastic5]"],
        "elastic2-aws": ["elastic2-doc-manager[elastic2,aws]"],
    },
    setup_requires=[
        "setuptools_scm>=1.5",
    ],
    python_requires=">=3.4",
)



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A set of utilities used throughout the mongo-connector
"""

import logging
import sys
import time

from bson.timestamp import Timestamp

from pymongo import errors

LOG = logging.getLogger(__name__)


def exception_wrapper(mapping):
    def decorator(f):
        def wrapped(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception:
                exc_type, exc_value, exc_tb = sys.exc_info()
                new_type = None
                for src_type in mapping:
                    if issubclass(exc_type, src_type):
                        new_type = mapping[src_type]
                        break

                if new_type is None:
                    raise
                raise new_type(str(exc_value)).with_traceback(exc_tb)

        return wrapped

    return decorator


def bson_ts_to_long(timestamp):
    """Convert BSON timestamp into integer.

    Conversion rule is based from the specs
    (http://bsonspec.org/#/specification).
    """
    return (timestamp.time << 32) + timestamp.inc


def long_to_bson_ts(val):
    """Convert integer into BSON timestamp.
    """
    seconds = val >> 32
    increment = val & 0xFFFFFFFF

    return Timestamp(seconds, increment)


def retry_until_ok(func, *args, **kwargs):
    """Retry code block until it succeeds.

    If it does not succeed in 120 attempts, the function re-raises any
    error the function raised on its last attempt.

    """
    max_tries = 120
    for i in range(max_tries):
        try:
            return func(*args, **kwargs)
        except RuntimeError:
            # Do not mask RuntimeError.
            raise
        except errors.OperationFailure as exc:
            if exc.code == 13 or (  # MongoDB >= 2.6 sets the error code,
                exc.details
                and "unauthorized" == exc.details.get("errmsg")  # MongoDB 2.4 does not.
            ):
                # Do not mask authorization failures.
                raise
            if i == max_tries - 1:
                LOG.exception(
                    "Call to %s failed too many times in " "retry_until_ok", func
                )
                raise
        except Exception:
            if i == max_tries - 1:
                LOG.exception(
                    "Call to %s failed too many times in " "retry_until_ok", func
                )
                raise
        time.sleep(1)


def log_fatal_exceptions(func):
    def wrapped(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception:
            LOG.exception("Fatal Exception")
            raise

    return wrapped



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import itertools
import time
import os
import sys

import pymongo
import requests


if sys.version_info[0] == 3:
    unicode = str

# Configurable hosts and ports used in the tests
db_user = unicode(os.environ.get("DB_USER", ""))
db_password = unicode(os.environ.get("DB_PASSWORD", ""))
# Extra keyword options to provide to Connector.
connector_opts = {}
if db_user:
    connector_opts = {"auth_username": db_user, "auth_key": db_password}

# Document count for stress tests
STRESS_COUNT = 100

# Test namespace, timestamp arguments
TESTARGS = ("test.test", 1)

_mo_address = os.environ.get("MO_ADDRESS", "localhost:8889")
_mongo_start_port = int(os.environ.get("MONGO_PORT", 27017))
_free_port = itertools.count(_mongo_start_port)

DEFAULT_OPTIONS = {"logappend": True, "setParameter": {"enableTestCommands": 1}}


_post_request_template = {}
if db_user and db_password:
    _post_request_template = {"login": db_user, "password": db_password}


def _mo_url(resource, *args):
    return "http://" + "/".join([_mo_address, resource] + list(args))


@atexit.register
def kill_all():
    clusters = requests.get(_mo_url("sharded_clusters")).json()
    repl_sets = requests.get(_mo_url("replica_sets")).json()
    servers = requests.get(_mo_url("servers")).json()
    for cluster in clusters["sharded_clusters"]:
        requests.delete(_mo_url("sharded_clusters", cluster["id"]))
    for rs in repl_sets["replica_sets"]:
        requests.delete(_mo_url("relica_sets", rs["id"]))
    for server in servers["servers"]:
        requests.delete(_mo_url("servers", server["id"]))


class MCTestObject(object):
    def proc_params(self):
        params = DEFAULT_OPTIONS.copy()
        params.update(self._proc_params)
        params["port"] = next(_free_port)
        return params

    def get_config(self):
        raise NotImplementedError

    def _make_post_request(self):
        config = _post_request_template.copy()
        config.update(self.get_config())
        ret = requests.post(_mo_url(self._resource), timeout=None, json=config)
        if not ret.ok:
            raise RuntimeError("Error sending POST to cluster: %s" % (ret.text,))

        ret = ret.json()
        if type(ret) == list:  # Will return a list if an error occurred.
            raise RuntimeError("Error sending POST to cluster: %s" % (ret,))
        return ret

    def client(self, **kwargs):
        client = pymongo.MongoClient(self.uri, **kwargs)
        if db_user:
            client.admin.authenticate(db_user, db_password)
        return client

    def stop(self):
        requests.delete(_mo_url(self._resource, self.id))


class Server(MCTestObject):

    _resource = "servers"

    def __init__(self, id=None, uri=None, **kwargs):
        self.id = id
        self.uri = uri
        self._proc_params = kwargs

    def get_config(self):
        return {"name": "mongod", "procParams": self.proc_params()}

    def start(self):
        if self.id is None:
            response = self._make_post_request()
            self.id = response["id"]
            self.uri = response.get("mongodb_auth_uri", response["mongodb_uri"])
        else:
            requests.post(
                _mo_url("servers", self.id), timeout=None, json={"action": "start"}
            )
        return self

    def stop(self, destroy=True):
        if destroy:
            super(Server, self).stop()
        else:
            requests.post(
                _mo_url("servers", self.id), timeout=None, json={"action": "stop"}
            )


class ReplicaSet(MCTestObject):

    _resource = "replica_sets"

    def __init__(self, id=None, uri=None, primary=None, secondary=None, **kwargs):
        self.id = id
        self.uri = uri
        self.primary = primary
        self.secondary = secondary
        self._proc_params = kwargs

    def get_config(self):
        return {
            "members": [
                {"procParams": self.proc_params()},
                {"procParams": self.proc_params()},
                {"rsParams": {"arbiterOnly": True}, "procParams": self.proc_params()},
            ]
        }

    def _init_from_response(self, response):
        self.id = response["id"]
        self.uri = response.get("mongodb_auth_uri", response["mongodb_uri"])
        for member in response["members"]:
            if member["state"] == 1:
                self.primary = Server(member["server_id"], member["host"])
            elif member["state"] == 2:
                self.secondary = Server(member["server_id"], member["host"])
        return self

    def start(self):
        # We never need to restart a replica set, only start new ones.
        return self._init_from_response(self._make_post_request())


class ReplicaSetSingle(ReplicaSet):
    def get_config(self):
        return {"members": [{"procParams": self.proc_params()}]}


class ShardedCluster(MCTestObject):

    _resource = "sharded_clusters"
    _shard_type = ReplicaSet

    def __init__(self, **kwargs):
        self.id = None
        self.uri = None
        self.shards = []
        self._proc_params = kwargs

    def get_config(self):
        return {
            "shards": [
                {"id": "demo-set-0", "shardParams": self._shard_type().get_config()},
                {"id": "demo-set-1", "shardParams": self._shard_type().get_config()},
            ],
            "routers": [self.proc_params()],
            "configsvrs": [self.proc_params()],
        }

    def start(self):
        # We never need to restart a sharded cluster, only start new ones.
        response = self._make_post_request()
        for shard in response["shards"]:
            if shard["id"] == "demo-set-0":
                repl1_id = shard["_id"]
            elif shard["id"] == "demo-set-1":
                repl2_id = shard["_id"]
        shard1 = requests.get(_mo_url("replica_sets", repl1_id)).json()
        shard2 = requests.get(_mo_url("replica_sets", repl2_id)).json()
        self.id = response["id"]
        self.uri = response.get("mongodb_auth_uri", response["mongodb_uri"])
        self.shards = [
            self._shard_type()._init_from_response(resp) for resp in (shard1, shard2)
        ]
        return self


class ShardedClusterSingle(ShardedCluster):
    _shard_type = ReplicaSetSingle


class MockGridFSFile:
    def __init__(self, doc, data):
        self._id = doc["_id"]
        self.filename = doc["filename"]
        self.upload_date = doc["upload_date"]
        self.md5 = doc["md5"]
        self.data = data
        self.length = len(self.data)
        self.pos = 0

    def get_metadata(self):
        return {
            "_id": self._id,
            "filename": self.filename,
            "upload_date": self.upload_date,
            "md5": self.md5,
        }

    def __len__(self):
        return self.length

    def read(self, n=-1):
        if n < 0 or self.pos + n > self.length:
            n = self.length - self.pos
        s = self.data[self.pos : self.pos + n]
        self.pos += n
        return s


def wait_for(condition, max_tries=60):
    """Wait for a condition to be true up to a maximum number of tries
    """
    cond = False
    while not cond and max_tries > 1:
        try:
            cond = condition()
        except Exception:
            pass
        time.sleep(1)
        max_tries -= 1
    return condition()


def assert_soon(condition, message=None, max_tries=60):
    """Assert that a condition eventually evaluates to True after at most
    max_tries number of attempts

    """
    if not wait_for(condition, max_tries=max_tries):
        raise AssertionError(message or "")


def close_client(client):
    if hasattr(type(client), "_process_kill_cursors_queue"):
        client._process_kill_cursors_queue()
        time.sleep(1)  # Wait for queue to clear.
    client.close()
